# Copyright 2019 Vitaliy Zakaznikov (TestFlows Test Framework http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
import sys
import time
import threading

import testflows.settings as settings

from .serialize import dumps
from .constants import id_sep, end_of_message
from .exceptions import exception as get_exception
from .message import Message
from .objects import Tag
from . import __version__

class TestOutput(object):
    """Test output protocol.

    :param io: message IO
    """
    rstrip_nulls = re.compile(r'(,null)+$')
    protocol_version = "TFSPv1.3"

    def __init__(self, test, io):
        self.io = io
        self.test = test
        self.msg_hash = ""
        self.msg_count = 0
        self.prefix = dumps([
            int(self.test.type),
            int(self.test.subtype),
            id_sep + id_sep.join(str(n) for n in self.test.id),
            int(self.test.flags),
            int(self.test.cflags)
        ])[1:-1]

    def message(self, keyword, message, rtime=None, stream=None):
        """Output message.

        :param keyword: keyword
        :param message: message
        :param rtime: relative time, default: None
        """
        rtime = rtime
        if rtime is None:
            rtime = round(time.time() - self.test.start_time, settings.time_resolution)
        msg = f"{self.msg_count},{self.prefix},{dumps(stream)},{rtime:.{settings.time_resolution}f},{str(message)}{end_of_message}"
        msg = self.rstrip_nulls.sub("", msg)
        msg_hash = settings.hash_func(f"{self.msg_hash},{keyword},{msg}".encode("utf-8")).hexdigest()[:settings.hash_length]
        self.msg_count += 1
        self.msg_hash = msg_hash
        self.io.write(f"{keyword},\"{msg_hash}\",{msg}")

    def protocol(self):
        """Output protocol version message.
        """
        msg = dumps(str(self.protocol_version))
        self.message(Message.PROTOCOL, msg, rtime=round(self.test.start_time, settings.time_resolution))

    def version(self):
        """Output framework version message.
        """
        msg = dumps(str(__version__))
        self.message(Message.VERSION, msg, rtime=round(self.test.start_time, settings.time_resolution))

    def input(self, message):
        """Output input message.

        :param message: message
        """
        msg = dumps(str(message))
        self.message(Message.INPUT, msg)

    def exception(self):
        """Output exception message.

        Note: must be called from within finally block
        """
        msg = dumps(get_exception())
        self.message(Message.EXCEPTION, msg)

    def test_message(self):
        """Output test message.

        :param test: test object
        """

        def object_fields(obj):
            return [getattr(obj, field) for field in obj._fields]

        def rstrip_list(l, value=[None, []]):
            """Remove all the items matching the value
            from the end of the list
            """
            while True:
                if l and l[-1] in value:
                    l.pop(-1)
                    continue
                break
            return l

        def str_or_repr(v):
            try:
                return str(v)
            except:
                return repr(v)

        msg = dumps(rstrip_list([
            self.test.name,
            round(self.test.start_time, settings.time_resolution),
            str(self.test.flags) or None,
            str(self.test.uid or "") or None,
            str(self.test.description or "") or None,
            [rstrip_list(object_fields(attr)) for attr in self.test.attributes],
            [rstrip_list(object_fields(req)) for req in self.test.requirements],
            [[str_or_repr(a) for a in rstrip_list(object_fields(arg))] for arg in self.test.args.values()],
            [rstrip_list(object_fields(Tag(tag))) for tag in self.test.tags],
            [rstrip_list(object_fields(user)) for user in self.test.users],
            [rstrip_list(object_fields(ticket)) for ticket in self.test.tickets],
            rstrip_list(object_fields(self.test.examples)) if self.test.examples else None,
        ]))[1:-1]
        self.message(Message.TEST, msg, rtime=0)

    def value(self, name, value):
        """Output value message.

        :param name: name
        :param value: value
        """
        msg = dumps([Message.VALUE, name, repr(value)])[1:-1]
        self.message(msg)

    def result(self, result):
        """Output result message.

        :param result: result object
        """
        msg = dumps([self.test.name, result.message, result.reason])[1:-1]
        self.message(getattr(Message, result.__class__.__name__.upper()), msg)

    def note(self, message):
        """Output note message.

        :param message: message
        """
        msg = dumps(str(message))
        self.message(Message.NOTE, msg)

    def debug(self, message):
        """Output debug message.

        :param message: message
        """
        msg = dumps(str(message))
        self.message(Message.DEBUG, msg)

    def trace(self, message):
        """Output trace message.

        :param message: message
        """
        msg = dumps(str(message))
        self.message(Message.TRACE, msg)


class TestInput(object):
    """Test input.
    """

    def __init__(self, test, io):
        self.test = test
        self.io = io


class TestIO(object):
    """Test input and output protocol.
    """

    def __init__(self, test):
        self.io = MessageIO(LogIO())
        self.output = TestOutput(test, self.io)
        self.input = TestInput(test, self.io)

    def message_io(self, name=None):
        """Return named line buffered message io.

        :param name: name of the message stream
        """
        return NamedMessageIO(self, name=name)

    def read(self, topic, timeout=None):
        """Read message.

        :param topic: message topic
        :param timeout: timeout, default: ``None``
        """
        raise NotImplementedError

    def write(self, msg, stream=None):
        """Write line buffered message.

        :param msg: line buffered message
        :param stream: name of the stream
        """
        if not msg:
            return
        self.output.message(Message.NONE, dumps(str(msg).rstrip()), stream=stream)

    def flush(self):
        self.io.flush()

    def close(self):
        self.io.close()

class MessageIO(object):
    """Message input and output.

    :param io: io stream to write and read
    """

    def __init__(self, io):
        self.io = io
        self.buffer = ""

    def read(self, topic, timeout=None):
        """Read message.

        :param topic: message topic
        :param timeout: timeout, default: ``None``
        """
        raise NotImplementedError

    def write(self, msg):
        """Write message.

        :param msg: message
        """
        if not msg:
            return
        if not "\n" in msg:
            self.buffer += msg
        else:
            self.buffer += msg
            messages = self.buffer.split("\n")
            # last message is incomplete
            for message in messages[:-1]:
                self.io.write(f"{message}\n")
            self.buffer = messages[-1]

    def flush(self):
        """Flush output buffer.
        """
        if self.buffer:
            self.io.write(f"{self.buffer}\n")
        self.buffer = ""

    def close(self):
        self.io.close()

class NamedMessageIO(MessageIO):
    """Message input and output.

    :param io: io stream to write and read
    :param name: name of the stream, default: None
    """

    def __init__(self, io, name=None):
        self.io = io
        self.buffer = ""
        self.stream = name

    def write(self, msg):
        """Write message.

        :param msg: message
        """
        if not msg:
            return
        if not "\n" in msg:
            self.buffer += msg
        else:
            self.buffer += msg
            messages = self.buffer.split("\n")
            # last message is incomplete
            for message in messages[:-1]:
                self.io.write(f"{message}\n", stream=self.stream)
            self.buffer = messages[-1]

    def flush(self):
        """Flush output buffer.
        """
        if self.buffer:
            self.io.write(f"{self.buffer}\n", stream=self.stream)
        self.buffer = ""


class LogReader(object):
    '''Read messages from the log.
    '''
    def __init__(self):
        self.fd = open(settings.read_logfile, "r", buffering=1, encoding="utf-8")

    def tell(self):
        return self.fd.tell()

    def seek(self, pos):
        return self.fd.seek(pos)

    def read(self, topic, timeout=None):
        raise NotImplementedError

    def close(self):
        self.fd.close()


class LogWriter(object):
    '''Singleton log file writer.
    '''
    lock = threading.Lock()
    instance = None

    def __new__(cls, *args, **kwargs):
        if not LogWriter.instance:
            LogWriter.instance = object.__new__(LogWriter)
        return LogWriter.instance

    def __init__(self):
        self.fd = open(settings.write_logfile, "a", buffering=1, encoding="utf-8")
        self.lock = threading.Lock()

    def write(self, msg):
        '''Write line buffered messages.

        :param msg: line buffered messages
        '''
        with self.lock:
            n = self.fd.write(msg)
            self.fd.flush()
            return n

    def flush(self):
        with self.lock:
            return self.fd.flush()

    def close(self):
        pass

class LogIO(object):
    '''Log file reader and writer.

    :param read: file descriptor for read
    :param write: file descriptor for write
    '''
    def __init__(self):
        self.writer = LogWriter()
        self.reader = LogReader()

    def write(self, msg):
        return self.writer.write(msg)

    def flush(self):
        return self.writer.flush()

    def tell(self):
        return self.reader.tell()

    def seek(self, pos):
        return self.reader.seek(pos)

    def read(self, topic, timeout=None):
        return self.reader.read(topic, timeout)

    def close(self):
        self.writer.close()
        self.reader.close()




# Copyright 2019 Vitaliy Zakaznikov (TestFlows Test Framework http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
import textwrap

import testflows.settings as settings

from testflows._core.flags import Flags, SKIP
from testflows._core.testtype import TestType, TestSubType
from testflows._core.transform.log import message
from testflows._core.objects import ExamplesTable
from testflows._core.utils.timefuncs import strftime, strftimedelta
from testflows._core.utils.timefuncs import localfromtimestamp
from testflows._core.name import split, basename, parentname
from testflows._core.cli.colors import color, cursor_up

strip_nones = re.compile(r'( None)+$')
indent = " " * 2
#: map of tests by name
tests_by_name = {}
#: map of tests by parent
tests_by_parent = {}

def color_keyword(keyword):
    return color(split(keyword)[-1], "white", attrs=["bold"])

def color_secondary_keyword(keyword):
    return color(split(keyword)[-1], "white", attrs=["bold", "dim"])

def color_other(other):
    return color(other, "white", attrs=["dim"])

def color_result(prefix, result):
    if result.startswith("X"):
        return color(prefix + result, "blue", attrs=["bold"])
    elif result.endswith("OK"):
        return color(prefix + result, "green", attrs=["bold"])
    elif result.endswith("Skip"):
        return color(prefix + result, "cyan", attrs=["bold"])
    # Error, Fail, Null
    return color(prefix + result, "red", attrs=["bold"])

def format_input(msg, keyword):
    flags = Flags(msg.p_flags)
    if flags & SKIP and settings.show_skipped is False:
        return
    out = color_other(f"{strftimedelta(msg.p_time):>20}{'':3}{indent * (msg.p_id.count('/') - 1)}{keyword}")
    out += color("\u270b " + msg.message, "yellow", attrs=["bold"]) + cursor_up() + "\n"
    return out

def format_description(msg, indent):
    first, rest = (msg.description.description.rstrip() + "\n").split("\n", 1)
    first = first.strip()
    if first:
        first += "\n"
    desc = f"{first}{textwrap.dedent(rest.rstrip())}".rstrip()
    desc = textwrap.indent(desc, indent + "  ")
    desc = color(desc, "white", attrs=["dim"])
    return desc + "\n"

def format_requirements(msg, indent):
    out = [f"{indent}{' ' * 2}{color_secondary_keyword('Requirements')}"]
    for req in msg.requirements:
        out.append(color(f"{indent}{' ' * 4}{req.name}", "white", attrs=["dim"]))
        out.append(color(f"{indent}{' ' * 6}version {req.version}", "white", attrs=["dim"]))
    return "\n".join(out) + "\n"

def format_attributes(msg, indent):
    out = [f"{indent}{' ' * 2}{color_secondary_keyword('Attributes')}"]
    for attr in msg.attributes:
        out.append(color(f"{indent}{' ' * 4}{attr.name}", "white", attrs=["dim"]))
        out.append(color(f"{indent}{' ' * 6}{attr.value}", "white", attrs=["dim"]))
    return "\n".join(out) + "\n"

def format_tags(msg, indent):
    out = [f"{indent}{' ' * 2}{color_secondary_keyword('Tags')}"]
    for tag in msg.tags:
        out.append(color(f"{indent}{' ' * 4}{tag.value}", "white", attrs=["dim"]))
    return "\n".join(out) + "\n"

def format_examples(msg, indent):
    examples = ExamplesTable(*msg.examples)
    out = [f"{indent}{' ' * 2}{color_secondary_keyword('Examples')}"]
    out.append(color(textwrap.indent(f"{examples}", prefix=f"{indent}{' ' * 4}"), "white", attrs=["dim"]))
    return "\n".join(out) + "\n"

def format_arguments(msg, indent):
    out = [f"{indent}{' ' * 2}{color_secondary_keyword('Arguments')}"]
    for arg in msg.args:
        out.append(color(f"{indent}{' ' * 4}{arg.name}", "white", attrs=["dim"]))
        out.append(color(textwrap.indent(f"{arg.value}", prefix=f"{indent}{' ' * 6}"), "white", attrs=["dim"]))
    return "\n".join(out) + "\n"

def format_users(msg, indent):
    out = [f"{indent}{' ' * 2}{color_secondary_keyword('Users')}"]
    for user in msg.users:
        out.append(color(f"{indent}{' ' * 4}{user.name}", "white", attrs=["dim"]))
    return "\n".join(out) + "\n"

def format_tickets(msg, indent):
    out = [f"{indent}{' ' * 2}{color_secondary_keyword('Tickets')}"]
    for ticket in msg.tickets:
        out.append(color(f"{indent}{' ' * 4}{ticket.name}", "white", attrs=["dim"]))
    return "\n".join(out) + "\n"

def and_keyword(msg, parent_name, keyword, subtype):
    """Handle processing of Given, When, Then, But, By and Finally
    keywords and convert them to And when necessary.
    """
    prev = tests_by_parent[parent_name][-2] if len(tests_by_parent.get(parent_name, [])) > 1 else None
    if prev and prev.p_subtype == subtype and tests_by_parent.get(prev.p_id) is None:
        keyword = "And"
    parent = tests_by_name.get(parent_name)
    if parent and parent.p_subtype == subtype and len(tests_by_parent.get(parent_name, [])) == 1:
        keyword = "And"
    return keyword

def format_test(msg, keyword):
    flags = Flags(msg.p_flags)
    if flags & SKIP and settings.show_skipped is False:
        return

    # add test to the tests map
    parent = parentname(msg.p_id)
    if tests_by_parent.get(parent) is None:
        tests_by_parent[parent] = []
    tests_by_parent[parent].append(msg)
    tests_by_name[msg.p_id] = msg

    if msg.p_type == TestType.Module:
        keyword += "Module"
    elif msg.p_type == TestType.Suite:
        keyword += "Suite"
    elif msg.p_type == TestType.Iteration:
        keyword += "Iteration"
    elif msg.p_type == TestType.Step:
        if msg.p_subtype == TestSubType.And:
            keyword += "And"
        elif msg.p_subtype == TestSubType.Given:
            keyword += and_keyword(msg, parent, "Given", TestSubType.Given)
        elif msg.p_subtype == TestSubType.When:
            keyword += and_keyword(msg, parent, "When", TestSubType.When)
        elif msg.p_subtype == TestSubType.Then:
            keyword += and_keyword(msg, parent, "Then", TestSubType.Then)
        elif msg.p_subtype == TestSubType.By:
            keyword += and_keyword(msg, parent, "By", TestSubType.By)
        elif msg.p_subtype == TestSubType.But:
            keyword += and_keyword(msg, parent, "But", TestSubType.But)
        elif msg.p_subtype == TestSubType.Finally:
            keyword += and_keyword(msg, parent, "Finally", TestSubType.Finally)
        else:
            keyword += "Step"
    else:
        if msg.p_subtype == TestSubType.Feature:
            keyword += "Feature"
        elif msg.p_subtype == TestSubType.Scenario:
            keyword += "Scenario"
        elif msg.p_subtype == TestSubType.Background:
            keyword += "Background"
        else:
            keyword += "Test"

    started = strftime(localfromtimestamp(msg.started))
    _keyword = color_keyword(keyword)
    _name = color_other(split(msg.name)[-1])
    _indent = f"{started:>20}" + f"{'':3}{indent * (msg.p_id.count('/') - 1)}"
    out = f"{color_other(_indent)}{_keyword} {_name}{color_other(', flags:' + str(flags) if flags else '')}\n"
    # convert indent to just spaces
    _indent = (len(_indent) + 3) * " "
    if msg.description:
        out += format_description(msg, _indent)
    if msg.tags:
        out += format_tags(msg, _indent)
    if msg.requirements:
        out += format_requirements(msg, _indent)
    if msg.attributes:
        out += format_attributes(msg, _indent)
    if msg.users:
        out += format_users(msg, _indent)
    if msg.tickets:
        out += format_tickets(msg, _indent)
    if msg.args:
        out += format_arguments(msg, _indent)
    if msg.examples:
        out += format_examples(msg, _indent)
    return out

def format_result(msg, prefix, result):
    if Flags(msg.p_flags) & SKIP and settings.show_skipped is False:
        return
    _result = color_result(prefix, result)
    _test = color_other(basename(msg.test))

    return (color_other(f"{strftimedelta(msg.p_time):>20}") +
        f"{'':3}{indent * (msg.p_id.count('/') - 1)}{_result} "
        f"{_test}{color_other(', ' + msg.test)}{color_other((', ' + msg.message) if msg.message else '')}"
        f"{color_other((', ' + msg.reason) if msg.reason else '')}\n")

def format_other(msg, keyword):
    if Flags(msg.p_flags) & SKIP and settings.show_skipped is False:
        return
    fields = ' '.join([str(f) for f in msg[message.Prefix.time + 1:]])
    if msg.p_stream:
        fields = f"[{msg.p_stream}] {fields}"
    fields = strip_nones.sub("", fields)

    fields = textwrap.indent(fields, prefix=(indent * (msg.p_id.count('/') - 1) + " " * 30))
    fields = fields.lstrip(" ")

    return color_other(f"{strftimedelta(msg.p_time):>20}{'':3}{indent * (msg.p_id.count('/') - 1)}{keyword} {fields}\n")

mark = "\u27e5"
result_mark = "\u27e5\u27e4"

formatters = {
    message.RawInput: (format_input, f"{mark} "),
    message.RawTest: (format_test, f"{mark}  "),
    message.RawDescription: (format_other, f"{mark}    :"),
    message.RawArgument: (format_other, f"{mark}    @"),
    message.RawAttribute: (format_other, f"{mark}    -"),
    message.RawRequirement: (format_other, f"{mark}    ?"),
    message.RawValue: (format_other, f"{mark}    ="),
    message.RawException: (format_other, f"{mark}    Exception:"),
    message.RawNote: (format_other, f"{mark}    [note]"),
    message.RawDebug: (format_other, f"{mark}    [debug]"),
    message.RawTrace: (format_other, f"{mark}    [trace]"),
    message.RawNone: (format_other, "    "),
    message.RawResultOK: (format_result, f"{result_mark} ", "OK"),
    message.RawResultFail: (format_result, f"{result_mark} ", "Fail"),
    message.RawResultError: (format_result, f"{result_mark} ", "Error"),
    message.RawResultSkip: (format_result, f"{result_mark} ", "Skip"),
    message.RawResultNull: (format_result, f"{result_mark} ", "Null"),
    message.RawResultXOK: (format_result, f"{result_mark} ", "XOK"),
    message.RawResultXFail: (format_result, f"{result_mark} ", "XFail"),
    message.RawResultXError: (format_result, f"{result_mark} ", "XError"),
    message.RawResultXNull: (format_result, f"{result_mark} ", "XNull")
}

def transform(stop):
    """Transform parsed log line into a nice format.

    :param stop: stop event
    """
    line = None

    while True:
        if line is not None:
            msg = line
            formatter = formatters.get(type(line), None)
            if formatter:
                line = formatter[0](line, *formatter[1:])
            else:
                line = None

        line = yield line



# Copyright 2019 Vitaliy Zakaznikov (TestFlows Test Framework http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time

from testflows._core.cli.colors import color
from testflows._core.utils.timefuncs import localfromtimestamp
from testflows._core.transform.log import message

def transform(stop):
    """Transform parsed log line into a nice format.

    :param stop: stop event
    """
    line = None
    divider = ""
    version = None
    started = None
    while True:
        if type(line) == message.RawVersion:
            version = line.version
            started = localfromtimestamp(line.p_time)
        if stop.is_set():
            if started is not None and version is not None:
                line = color(f"{divider}\nExecuted on {started:%b %d,%Y %-H:%M}\nTestFlows Test Framework v{version}\n",
                    "white", attrs=["dim"])
            else:
                line = None
        else:
            line = None
        line = yield line


# Copyright 2019 Vitaliy Zakaznikov (TestFlows Test Framework http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from testflows._core.test import TestBase
from testflows._core.test import Module, Suite, Test, Step, NullStep, run
from testflows._core.test import TestCase, TestSuite, TestModule
from testflows._core.test import Attributes, Requirements, Users, Tickets, Examples
from testflows._core.test import Name, Description, Uid, Tags, TestClass
from testflows._core.test import Feature, Background, Scenario
from testflows._core.test import Given, When, Then, And, But, By, Finally
from testflows._core.test import Steps
from testflows._core.test import TestFeature, TestScenario
from testflows._core.funcs import *
from testflows._core.filters import the
from testflows._core.objects import *
from testflows._core.name import *
from testflows._core.flags import *
from testflows._core import __author__, __version__, __license__

import testflows._core.utils as utils



# Copyright 2019 Vitaliy Zakaznikov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import importlib

from json import JSONEncoder

from .baseobject import TestArg

class Encoder(JSONEncoder):
    """Argument value encoder.
    """
    def default(self, o):
        if isinstance(o, TestArg):
            objtype = ".".join([o.__class__.__module__, o.__class__.__name__])
            args = list(o.initargs.args)
            args.append(o.initargs.kwargs)
            return {"@py:%s" % objtype : args}
        return super(Encoder, self).default(o)

def object_hook(o):
    if not o:
        return o
    key = next(iter(o))
    if key.startswith("@py:"):
        initargs = o.get(key)
        objtype = key[4:]
        return decode_argument(objtype, initargs)
    return o
    
def decode_argument(objtype, initargs):
    module, cls = objtype.rsplit('.',1)
    args = initargs[:-1] 
    kwargs = initargs[-1]
    return getattr(importlib.import_module(module), cls)(*args, **kwargs)

def dumps(o, *args, **kwargs):
    """Serialize argument value to a JSON formatted string.
    """
    cls = kwargs.pop('cls', Encoder)
    return json.dumps(o, cls=cls, separators=(',', ':'), *args, **kwargs)

def loads(s, *args, **kwargs):
    """Deserializes JSON string to argument value.
    """
    hook = kwargs.pop('object_hook', object_hook)
    return json.loads(s, object_hook=hook, *args, **kwargs)



# Copyright 2019 Vitaliy Zakaznikov (TestFlows Test Framework http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
class TestType(object):
    """Test type."""
    Module = 40
    Suite = 30
    Test = 20
    Iteration = 15
    Step = 10

class TestSubType(object):
    """Test behaviour subtype."""
    Feature = 60
    Scenario = 50
    Background = 40
    Given = 30
    When = 20
    Then = 10
    And = 8
    But = 7
    By = 6
    Finally = 5
    Empty = 0




# Copyright 2019 Vitaliy Zakaznikov (TestFlows Test Framework http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from enum import IntEnum
from collections import namedtuple

class Message(IntEnum):
    NONE = 0
    TEST = 1
    NULL = 2
    OK = 3
    FAIL = 4
    SKIP = 5
    ERROR = 6
    EXCEPTION = 7
    VALUE = 8
    NOTE = 9
    DEBUG = 10
    TRACE = 11
    XOK = 12
    XFAIL = 13
    XERROR = 14
    XNULL = 15
    PROTOCOL = 16
    INPUT = 17
    VERSION = 18

ResultMessages = [
    Message.OK,
    Message.FAIL,
    Message.SKIP,
    Message.NULL,
    Message.ERROR,
    Message.XOK,
    Message.XFAIL,
    Message.XERROR,
    Message.XNULL
]

MessageMap = namedtuple(
        "MessageMap",
        "NONE TEST NULL OK FAIL SKIP ERROR "
        "EXCEPTION VALUE NOTE DEBUG TRACE "
        "XOK XFAIL XERROR XNULL PROTOCOL INPUT "
        "VERSION"
    )


# Copyright 2019 Vitaliy Zakaznikov (TestFlows Test Framework http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
def transform(file, tail=False, offset=False):
    """Read lines from a file-like object.

    :param file: open file handle
    :param tail: tail mode, default: False
    :param offset: include offset with the message, default: False
    """
    yield None
    line = ""
    pos = 0
    while True:
        line += file.readline()
        if line.endswith("\n"):
            if offset:
                yield (line, pos)
                pos += len(line.encode("utf-8"))
            else:
                yield line
            line = ""



# Copyright 2019 Vitaliy Zakaznikov (TestFlows Test Framework http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
import testflows.settings as settings

from testflows._core.flags import Flags, SKIP
from testflows._core.testtype import TestType
from testflows._core.transform.log import message
from testflows._core.name import split
from testflows._core.cli.colors import color

indent = " " * 2

def color_result(result, attrs=None):
    if attrs is None:
        attrs = ["bold"]
    if result.startswith("X"):
        return functools.partial(color, color="blue", attrs=attrs)
    elif result == "OK":
        return functools.partial(color, color="green", attrs=attrs)
    elif result == "Skip":
        return functools.partial(color, color="cyan", attrs=attrs)
    # Error, Fail, Null
    elif result == "Error":
        return functools.partial(color, color="yellow", attrs=attrs)
    elif result == "Fail":
        return functools.partial(color, color="red", attrs=attrs)
    elif result == "Null":
        return functools.partial(color, color="magenta", attrs=attrs)
    else:
        raise ValueError(f"unknown result {result}")

def add_result(msg, results, result):
    if msg.p_type < TestType.Test:
        if not result.startswith("X"):
            return
    flags = Flags(msg.p_flags)
    if flags & SKIP and settings.show_skipped is False:
        return
    if not result in ("OK", "Skip"):
        results[msg.p_id] = (msg, result)

processors = {
    message.RawResultOK: (add_result, f"OK"),
    message.RawResultFail: (add_result, f"Fail"),
    message.RawResultError: (add_result, f"Error"),
    message.RawResultSkip: (add_result, f"Skip"),
    message.RawResultNull: (add_result, f"Null"),
    message.RawResultXOK: (add_result, f"XOK"),
    message.RawResultXFail: (add_result, f"XFail"),
    message.RawResultXError: (add_result, f"XError"),
    message.RawResultXNull: (add_result, f"XNull")
}

def generate(results):
    """Generate report"""
    if not results:
        return

    xfails = ""
    fails = ""

    for entry in results:
        msg, result = results[entry]
        _color = color_result(result)
        if not result.startswith("X"):
            continue
        xfails += _color('\u2718') + f" [ { _color(result) } ] {msg.test}\n"

    if xfails:
        xfails = color("\nKnown\n\n", "white", attrs=["bold"]) + xfails

    for entry in results:
        msg, result = results[entry]
        _color = color_result(result)
        if result.startswith("X"):
            continue
        fails += _color("\u2718") + f" [ {_color(result)} ] {msg.test}\n"
    if fails:
        fails = color("\nFailing\n\n", "white", attrs=["bold"]) + fails

    report = f"{xfails}{fails}"

    return report or None

def transform(stop):
    """Transform parsed log line into a short format.
    """
    line = None
    results = {}
    while True:
        if line is not None:
            processor = processors.get(type(line), None)
            if processor:
                processor[0](line, results, *processor[1:])
            if stop.is_set():
                line = generate(results)
            else:
                line = None
        line = yield line


# Copyright 2019 Vitaliy Zakaznikov (TestFlows Test Framework http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import textwrap

from argparse import ArgumentParser as ArgumentParserBase

from .common import epilog
from .common import description
from .common import HelpFormatter
from .handlers.transform.handler import Handler as transform_handler
from .handlers.document.handler import Handler as document_handler
from .handlers.requirement.handler import Handler as requirement_handler
from .handlers.report.handler import Handler as report_handler
from .handlers.show.handler import Handler as show_handler

from testflows._core import __version__, __license__

class ArgumentParser(ArgumentParserBase):
    """Customized argument parser.
    """
    def __init__(self, *args, **kwargs):
        description_prog = kwargs.pop("description_prog", None)
        kwargs["epilog"] = kwargs.pop("epilog", epilog())
        kwargs["description"] = description(textwrap.dedent(kwargs.pop("description", "")), prog=description_prog)
        kwargs["formatter_class"] = kwargs.pop("formatter_class", HelpFormatter)
        return super(ArgumentParser, self).__init__(*args, **kwargs)

parser = ArgumentParser(prog="tfs")

parser.add_argument("--debug", dest="debug", action="store_true",
                    help="enable debugging mode", default=False)
parser.add_argument("--no-colors", dest="no_colors", action="store_true",
                    help="disable terminal color highlighting", default=False)
parser.add_argument("--show-skipped", dest="show_skipped", action="store_true",
                    help="show skipped tests, default: False", default=False)
parser.add_argument("-v", "--version", action="version", version=f"{__version__}")
parser.add_argument("--license", action="version", help="show program's license and exit", version=f"{__license__}")

commands = parser.add_subparsers(title='commands', metavar='command', description=None, help=None)

report_handler.add_command(commands)
transform_handler.add_command(commands)
requirement_handler.add_command(commands)
document_handler.add_command(commands)
show_handler.add_command(commands)



# Copyright 2019 Vitaliy Zakaznikov (TestFlows Test Framework http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import re
import sys
import glob
import tempfile
import threading

import testflows.settings as settings

from .transform.log.pipeline import RawLogPipeline
from .transform.log.pipeline import NiceLogPipeline
from .transform.log.pipeline import DotsLogPipeline
from .transform.log.pipeline import ShortLogPipeline

def cleanup():
    """Clean up old temporary log files.
    """
    parser = re.compile(r".*testflows.(?P<pid>\d+).log")

    def pid_exists(pid):
        """Check if pid is alive on UNIX.

        :param pid: pid
        """
        # NOTE: pid == 0 returns True
        if pid < 0: return False
        try:
            os.kill(pid, 0)
        except ProcessLookupError:
            # errno.ESRCH - no such process
            return False
        except PermissionError:
            # errno.EPERM - operation not permitted (i.e., process exists)
            return True
        else:
            #  no error, we can send a signal to the process
            return True

    for file in glob.glob(os.path.join(tempfile.gettempdir(), "testflows.*.log")):
        match = parser.match(file)
        if not match:
            continue
        pid = int(match.groupdict()['pid'])
        if not pid_exists(pid):
            try:
                os.remove(file)
            except OSError:
                raise

def stdout_raw_handler():
    """Handler to output messages to sys.stdout
    using "raw" format.
    """
    with open(settings.read_logfile, "a+", buffering=1, encoding="utf-8") as log:
        log.seek(0)
        RawLogPipeline(log, sys.stdout, tail=True).run()

def stdout_short_handler():
    """Handler to output messages to sys.stdout
    using "short" format.
    """
    with open(settings.read_logfile, "a+", buffering=1, encoding="utf-8") as log:
        log.seek(0)
        ShortLogPipeline(log, sys.stdout, tail=True).run()

def stdout_nice_handler():
    """Handler to output messages to sys.stdout
    using "nice" format.
    """
    with open(settings.read_logfile, "a+", buffering=1, encoding="utf-8") as log:
        log.seek(0)
        NiceLogPipeline(log, sys.stdout, tail=True).run()

def stdout_dots_handler():
    """Handler to output messages to sys.stdout
    using "dots" format.
    """
    with open(settings.read_logfile, "a+", buffering=1, encoding="utf-8") as log:
        log.seek(0)
        DotsLogPipeline(log, sys.stdout, tail=True).run()

def stdout_silent_handler():
    """Handler that prints no output to sys.stdout.
    """
    pass

def init():
    """Initialization before we run the first test.
    """
    cleanup()

    output_handler_map = {
        "raw": stdout_raw_handler,
        "nice": stdout_nice_handler,
        "silent": stdout_silent_handler,
        "quiet": stdout_silent_handler,
        "short": stdout_short_handler,
        "dots": stdout_dots_handler,
    }
    # start stdout output handler
    handler = threading.Thread(target=output_handler_map[settings.output_format])
    handler.name = "tfs-output"
    handler.start()
    return True



# Copyright 2019 Vitaliy Zakaznikov (TestFlows Test Framework http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License
__all__ = ["ExitException", "ExitWithError", "ExitWithWarning", "ExitWithSuccess"]

class ExitException(Exception):
    """Base class for all exit exceptions.

    :param exitcode: exit code
    """
    def __init__(self, exitcode, message):
        self.exitcode = exitcode
        self.message = message
        super(ExitException, self).__init__(message)

class ExitWithError(ExitException):
    """Exit with error exception."""
    def __init__(self, message, exitcode=1):
        super(ExitWithError, self).__init__(exitcode, message)

class ExitWithWarning(ExitException):
    """Exit with warning exception."""
    def __init__(self, message, exitcode=0):
        super(ExitWithWarning, self).__init__(exitcode, message)

class ExitWithSuccess(ExitException):
    """Exit with success exception."""
    def __init__(self, message, exitcode=0):
        super(ExitWithSuccess, self).__init__(exitcode, message)


# Copyright 2019 Vitaliy Zakaznikov (TestFlows Test Framework http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import testflows.settings as settings

from testflows._core.flags import Flags, SKIP
from testflows._core.testtype import TestType, TestSubType
from testflows._core.transform.log import message
from testflows._core.utils.timefuncs import strftimedelta
from testflows._core.cli.colors import color

def color_line(line):
    return color(line, "white", attrs=["dim"])

def color_result(result, text):
    if result.startswith("X"):
        return color(text, "blue", attrs=["bold"])
    elif result == "OK":
        return color(text, "green", attrs=["bold"])
    elif result == "Skip":
        return color(text, "white", attrs=["dim"])
    # Error, Fail, Null
    elif result == "Error":
        return color(text, "yellow", attrs=["bold"])
    elif result == "Fail":
         return color(text, "red", attrs=["bold"])
    elif result == "Null":
        return color(text, "magenta", attrs=["bold"])
    else:
        raise ValueError(f"unknown result {result}")

class Counts(object):
    def __init__(self, name, units, ok, fail, skip, error, null, xok, xfail, xerror, xnull):
        self.name = name
        self.units = units
        self.ok = ok
        self.fail = fail
        self.skip = skip
        self.error = error
        self.null = null
        self.xok = xok
        self.xfail = xfail
        self.xerror = xerror
        self.xnull = xnull

    def __bool__(self):
        return self.units > 0

    def __data__(self):
        data = {}
        counts = {}
        data["units"] = self.units
        data["name"] = self.name if self.units != 1 else self.name.rstrip('s')
        data["counts"] = counts
        if self.ok > 0:
            counts["OK"] = self.ok
        if self.fail > 0:
            counts["Fail"] = self.fail
        if self.skip > 0:
            counts["Skip"] = self.skip
        if self.error > 0:
            counts["Error"] = self.error
        if self.null > 0:
            counts["Null"] = self.null
        if self.xok > 0:
            counts["XOK"] = self.xok
        if self.xfail > 0:
            counts["XFail"] = self.xfail
        if self.xerror > 0:
            counts["XError"] = self.xerror
        if self.xnull > 0:
            counts["XNull"] = self.xnull
        return data

    def __str__(self):
        s = f"{self.units} {self.name if self.units != 1 else self.name.rstrip('s')} ("
        s = color(s, "white", attrs=["bold"])
        r = []
        if self.ok > 0:
            r.append(color_result("OK", f"{self.ok} ok"))
        if self.fail > 0:
            r.append(color_result("Fail", f"{self.fail} failed"))
        if self.skip > 0:
            r.append(color_result("Skip", f"{self.skip} skipped"))
        if self.error > 0:
            r.append(color_result("Error", f"{self.error} errored"))
        if self.null > 0:
            r.append(color_result("Null", f"{self.null} null"))
        if self.xok > 0:
            r.append(color_result("XOK", f"{self.xok} xok"))
        if self.xfail > 0:
            r.append(color_result("XFail", f"{self.xfail} xfail"))
        if self.xerror > 0:
            r.append(color_result("XError", f"{self.xerror} xerror"))
        if self.xnull > 0:
            r.append(color_result("XNull", f"{self.xnull} xnull"))
        s += color(", ", "white", attrs=["bold"]).join(r)
        s += color(")\n", "white", attrs=["bold"])
        return s

def format_test(msg, counts):
    flags = Flags(msg.p_flags)
    if flags & SKIP and settings.show_skipped is False:
        return
    if msg.p_type == TestType.Module:
        counts["module"].units += 1
    elif msg.p_type == TestType.Suite:
        counts["suite"].units += 1
    elif msg.p_type == TestType.Iteration:
        counts["iteration"].units += 1
    elif msg.p_type == TestType.Step:
        counts["step"].units += 1
    else:
        if msg.p_subtype == TestSubType.Feature:
            counts["feature"].units += 1
        elif msg.p_subtype == TestSubType.Scenario:
            counts["scenario"].units += 1
        else:
            counts["test"].units += 1

def format_result(msg, counts):
    if Flags(msg.p_flags) & SKIP and settings.show_skipped is False:
        return

    _result_name_map = {
        message.RawResultOK: "ok",
        message.RawResultFail: "fail",
        message.RawResultNull: "null",
        message.RawResultError: "error",
        message.RawResultXOK: "xok",
        message.RawResultXFail: "xfail",
        message.RawResultXNull: "xnull",
        message.RawResultXError: "xerror",
        message.RawResultSkip: "skip"
    }

    _name = _result_name_map[type(msg)]

    if msg.p_type == TestType.Module:
        setattr(counts["module"], _name, getattr(counts["module"], _name) + 1)
    elif msg.p_type == TestType.Suite:
        setattr(counts["suite"], _name, getattr(counts["suite"], _name) + 1)
    elif msg.p_type == TestType.Iteration:
        setattr(counts["iteration"], _name, getattr(counts["iteration"], _name) + 1)
    elif msg.p_type == TestType.Step:
        setattr(counts["step"], _name, getattr(counts["step"], _name) + 1)
    else:
        if msg.p_subtype == TestSubType.Feature:
            setattr(counts["feature"], _name, getattr(counts["feature"], _name) + 1)
        elif msg.p_subtype == TestSubType.Scenario:
            setattr(counts["scenario"], _name, getattr(counts["scenario"], _name) + 1)
        else:
            setattr(counts["test"], _name, getattr(counts["test"], _name) + 1)

formatters = {
    message.RawTest: (format_test,),
    message.RawResultOK: (format_result,),
    message.RawResultFail: (format_result,),
    message.RawResultError: (format_result,),
    message.RawResultSkip: (format_result,),
    message.RawResultNull: (format_result,),
    message.RawResultXOK: (format_result,),
    message.RawResultXFail: (format_result,),
    message.RawResultXError: (format_result,),
    message.RawResultXNull: (format_result,)
}

def all_counts():
    return {
        "module": Counts("modules", *([0] * 10)),
        "suite": Counts("suites", *([0] * 10)),
        "test": Counts("tests", *([0] * 10)),
        "iteration": Counts("iterations", *([0] * 10)),
        "step": Counts("steps", *([0] * 10)),
        "feature": Counts("features", *([0] * 10)),
        "scenario": Counts("scenarios", *([0] * 10))
    }

def transform(stop):
    """Totals report.

    :param stop: stop event
    """
    counts = all_counts()
    line = None

    while True:
        if line is not None:
            msg = line
            formatter = formatters.get(type(line), None)
            if formatter:
                formatter[0](line, *formatter[1:], counts=counts)
            line = None

            if stop.is_set():
                if line is None:
                    line = ""
                line += "\n"
                line_icon = "" #"\u27a4 "
                if counts["module"]:
                    line += line_icon + str(counts["module"])
                if counts["suite"]:
                    line += line_icon + str(counts["suite"])
                if counts["test"]:
                    line += line_icon + str(counts["test"])
                if counts["feature"]:
                    line += line_icon + str(counts["feature"])
                if counts["scenario"]:
                    line += line_icon + str(counts["scenario"])
                if counts["iteration"]:
                    line += line_icon + str(counts["iteration"])
                if counts["step"]:
                    line += line_icon + str(counts["step"])
                line += color_line(f"\nTotal time {strftimedelta(msg.p_time)}\n")
        line = yield line


# Copyright 2019 Vitaliy Zakaznikov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .exceptions import RequirementError
from .baseobject import TestObject, TestArg, Table
from .baseobject import get, hash

class Result(TestObject):
    _fields = ("test", "message", "reason")
    _defaults = (None, None)

    def __init__(self, test, message=None, reason=None):
        self.test = test
        self.message = message
        self.reason = reason
        return super(Result, self).__init__()

    def xout(self, reason=None):
        return self

    def __bool__(cls):
        return True

    def __eq__(self, o):
        return type(self) == o
  
    def __ne__(self, o):
        return not self == o

class XResult(Result):
    pass

class OK(Result):
    def xout(self, reason):
        return XOK(self.test, self.message, reason)

class XOK(XResult):
    pass

class Fail(Result):
    def xout(self, reason):
        return XFail(self.test, self.message, reason)

    def __bool__(self):
        return False

class XFail(XResult):
    pass

class Skip(Result):
    pass

class Error(Result):
    def xout(self, reason):
        return XError(self.test, self.message, reason)

    def __bool__(self):
        return False

class XError(XResult):
    pass

class Null(Result):
    def xout(self, reason):
        return XNull(self.test, self.message, reason)

    def __bool__(self):
        return False

class XNull(XResult):
    pass

class Tag(TestObject):
    _fields = ("value",)
    _defaults = ()

    def __init__(self, value):
        self.value = value
        return super(Tag, self).__init__()

class Argument(TestObject):
    _fields = ("name", "value", "group", "type", "uid")
    _defaults = (None,) * 4
    uid = None
    type = None
    group = None

    def __init__(self, name, value=None, type=None, group=None, uid=None):
        self.name = name
        self.value = value
        self.type = get(type, self.type)
        self.group = get(group, self.group)
        self.uid = get(uid, self.uid)
        return super(Argument, self).__init__()

class Attribute(TestObject):
    _fields = ("name", "value", "type", "group", "uid")
    _defaults = (None,) * 3
    uid = None
    type = None
    group = None
    
    def __init__(self, name, value, type=None, group=None, uid=None):
        self.name = name
        self.value = value
        self.type = get(type, self.type)
        self.group = get(group, self.group)
        self.uid = get(uid, self.uid)
        return super(Attribute, self).__init__()

class Requirement(TestObject):
    _fields = ("name", "version", "description",
            "link", "priority", "type", "group", "uid")
    _defaults = (None,) * 6
    uid = None
    link = None
    priority = None
    type = None
    group = None
    
    def __init__(self, name, version, description=None, link=None,
            priority=None, type=None, group=None, uid=None):
        self.name = name
        self.version = version
        self.description = get(description, self.__doc__)
        self.link = get(link, self.link)
        self.priority = get(priority, self.priority)
        self.type = get(type, self.type)
        self.group = get(group, self.group)
        self.uid = get(uid, self.uid)
        return super(Requirement, self).__init__()  

    def __call__(self, *version):
        if not self.version in version:
            raise RequirementError("requirement version %s is not in %s" % (self.version, list(version)))
        return self

class Measurement(TestObject):
    _fields = ("name", "value", "units", "type", "group", "uid")
    _defaults = (None,) * 3
    uid = None
    type = None
    group = None
    
    def __init__(self, name, value, units, type=None, group=None, uid=None):
        self.name = name
        self.value = value
        self.units = units
        self.type = get(type, self.type)
        self.group = get(group, self.group)
        self.uid = get(uid, self.uid)
        return super(Measurement, self).__init__()

class Output(TestObject):
    _fields = ("name", "value", "type", "group", "uid")
    _defaults = (None,) * 3
    uid = None
    type = None
    group = None

    def __init__(self, name, value, type=None, group=None, uid=None):
        self.name = name
        self.value = value
        self.type = get(type, self.type)
        self.group = get(group, self.group)
        self.uid = get(uid, self.uid)
        return super(Output, self).__init__()

class Project(TestObject):
    _fields = ("name", "type", "group", "uid",
        "tags", "attributes")
    _defaults = (None,) * 5
    uid = None
    type = None
    group = None
    tags = []
    attributes = []
    
    def __init__(self, name, type=None, group=None, tags=None, attributes=None, uid=None):
        self.name = name
        self.type = get(type, self.type)
        self.group = get(group, self.group)
        self.tags = get(tags, self.tags)
        self.attributes = get(attributes, self.attributes)
        self.uid = get(uid, self.uid)
        return super(Project, self).__init__()

class Ticket(TestObject):
    _fields = ("name", "link", "type", "group", "uid")
    _defaults = (None,) * 4
    uid = None
    link = None
    type = None
    group = None
    
    def __init__(self, name, link=None, type=None, group=None, uid=None):
        self.name = name
        self.link = get(link, self.link)
        self.type = get(type, self.type)
        self.group = get(group, self.group)
        self.uid = get(uid, self.uid)
        return super(Ticket, self).__init__()

class User(TestObject):
    _fields = ("name", "type", "group", "link", "uid",
        "tags", "attributes")
    _defaults = (None,) * 6
    uid = None
    link = None
    type = None
    group = None
    tags = []
    attributes = []
    
    def __init__(self, name, link=None, type=None, group=None, tags=None, attributes=None, uid=None):
        self.name = name
        self.link = get(link, self.link)
        self.type = get(type, self.type)
        self.group = get(group, self.group)
        self.tags = get(tags, self.tags)
        self.attributes = get(attributes, self.attributes)
        self.uid = get(uid, self.uid)
        return super(User, self).__init__()

class Environment(TestObject):
    _fields = ("name", "type", "group", "uid",
        "tags", "attributes", "devices")
    _defaults = (None,) * 6
    uid = None
    type = None
    group = None
    tags = []
    attributes = []
    devices = []

    def __init__(self, name, type=None, group=None, tags=None, attributes=None,
            devices=None, uid=None):
        self.name = name
        self.type = get(type, self.type)
        self.group = get(group, self.group)
        self.tags = get(tags, self.tags)
        self.attributes = get(attributes, self.attributes)
        self.devices = get(devices, self.devices)
        self.uid = get(uid, self.uid)
        return super(User, self).__init__()

class Device(TestObject):
    _fields = ("name", "type", "group", "tags", "uid",
        "attributes", "requirements",
        "software", "hardware")
    _defaults = (None,) * 7
    uid = None
    type = None
    group = None
    tags = []
    attributes = []
    requirements = []
    software = []
    hardware = []
    
    def __init__(self, name, type=None, group=None, tags=None,
            attributes=None, requirements=None,
            software=None, hardware=None, uid=None):
        self.name = name
        self.type = get(type, self.type)
        self.group = get(group, self.group)
        self.tags = get(tags, self.tags)
        self.attributes = get(attributes, self.attributes)
        self.requirements = get(requirements, self.requirements)
        self.software = get(software, self.software)
        self.hardware = get(hardware, self.hardware)
        self.uid = get(uid, self.uid)
        return super(Device, self).__init__()

class Software(TestObject):
    _fields = ("name", "type", "group", "uid",
        "tags", "attributes", "requirements")
    _defaults = (None,) * 5
    uid = None
    type = None
    group = None
    tags = []
    attributes = []
    requirements = []

    def __init__(self, name, type=None, group=None, tags=None,
            attributes=None, requirements=None, uid=None):
        self.name = name
        self.type = get(type, self.type)
        self.group = get(group, self.group)
        self.tags = get(tags, self.tags)
        self.attributes = get(attributes, self.attributes)
        self.requirements = get(requirements, self.requirements)
        self.uid = get(uid, self.uid)
        return super(Software, self).__init__()

class Hardware(TestObject):
    _fields = ("name", "type", "group", "uid",
        "tags", "attributes", "requirements")
    _defaults = (None,) * 5
    uid = None
    type = None
    group = None
    tags = []
    attributes = []
    requirements = []

    def __init__(self, name, type=None, group=None, tags=None,
            attributes=None, requirements=None, uid=None):
        self.name = name
        self.type = get(type, self.type)
        self.group = get(group, self.group)
        self.tags = get(tags, self.tags)
        self.attributes = get(attributes, self.attributes)
        self.requirements = get(requirements, self.requirements)
        self.uid = get(uid, self.uid)
        return super(Hardware, self).__init__()

class Job(TestObject):
    _fields = ("name", "type", "group",
        "tags", "user",
        "attributes", "requirements", "uid")
    _defaults = (None,) * 5
    uid = None
    user = None
    type = None
    group = None
    tags = []
    attributes = []
    requirements = []
    
    def __init__(self, name, user, type=None, group=None, tags=None,
            attributes=None, requirements=None, uid=None):
        self.name = name
        self.user = user
        self.type = get(type, self.type)
        self.group = get(group, self.group)
        self.tags = get(tags, self.tags)
        self.attributes = get(attributes, self.attributes)
        self.requirements = get(requirements, self.requirements)
        self.uid = get(uid, self.uid)
        return super(Software, self).__init__()

class ExamplesTable(Table):
    _row_type_name = "Example"

    @classmethod
    def from_table(cls, table):
        """Create examples table from a table.
        """
        return cls(header=" ".join(table.row_type._fields), rows=table, row_format=table.row_format)



# Copyright 2019 Vitaliy Zakaznikov (TestFlows Test Framework http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json

import testflows.settings as settings

from testflows._core.transform.log import message
from testflows._core.message import Message
from testflows._core.constants import id_sep

def transform(stop=None):
    """Transform raw message into raw format.
    """
    msg = None
    stop_id = f"{id_sep}{settings.test_id}"
    prefix = message.RawFormat.prefix

    while True:
        if msg is not None:
            try:
                if stop:
                    fields = json.loads(f"[{msg}]")
                    if fields[prefix.keyword] \
                            in (Message.SKIP, Message.NULL, Message.FAIL,
                                Message.ERROR, Message.OK) \
                            and fields[prefix.id] == stop_id:
                        stop.set()
            except (IndexError, Exception):
                raise Exception(f"invalid message: {msg}\n")
        msg = yield msg



# Copyright 2019 Vitaliy Zakaznikov (TestFlows Test Framework http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import threading

from testflows._core.message import Message, ResultMessages
from .read import transform as read_transform
from .parse import transform as parse_transform
from .nice import transform as nice_transform
from .dots import transform as dots_transform
from .write import transform as write_transform
from .stop import transform as stop_transform
from .raw import transform as raw_transform
from .short import transform as short_transform
from .index import transform as index_transform
from .read_and_filter import transform as read_and_filter_transform
from .report.passing import transform as passing_report_transform
from .report.fails import transform as fails_report_transform
from .report.totals import transform as totals_report_transform
from .report.version import transform as version_report_transform
from .report.results import transform as results_transform

class Pipeline(object):
    """Combines multiple steps into a pipeline
    that can be executed.
    """
    def __init__(self, steps):
        self.steps = steps
        # start all the generators
        for step in self.steps:
            next(step)

    def run(self):
        """Execute pipeline.
        """
        item = None
        while True:
            try:
                for step in self.steps:
                    item = step.send(item)
                    if item is None:
                        break
            except StopIteration:
                break

def fanout(*steps):
    """Single step of pipeline
    that feeds the same input to
    multiple steps and produces
    a list of outputs from each step.

    :param *steps: fan out steps
    """
    item = None
    outputs = []
    while True:
        for step in steps:
            output = step.send(item)
            if output is not None:
                outputs.append(output)
        item = yield outputs or None
        outputs = []

def fanin(combinator):
    """Combine multiple outputs into one.
    using the combinator
    """
    item = None
    while True:
        if item is not None:
            item = combinator(item)
        item = yield item

class RawLogPipeline(Pipeline):
    def __init__(self, input, output, tail=False):
        stop_event = threading.Event()

        steps = [
            read_transform(input, tail=tail),
            raw_transform(stop_event),
            write_transform(output),
            stop_transform(stop_event)
        ]
        super(RawLogPipeline, self).__init__(steps)

class ShortLogPipeline(Pipeline):
    def __init__(self, input, output, tail=False):
        stop_event = threading.Event()

        steps = [
            read_transform(input, tail=tail),
            parse_transform(stop_event),
            fanout(
                short_transform(),
                passing_report_transform(stop_event),
                fails_report_transform(stop_event),
                totals_report_transform(stop_event),
                version_report_transform(stop_event),
            ),
            fanin(
                "".join
            ),
            write_transform(output),
            stop_transform(stop_event)
        ]
        super(ShortLogPipeline, self).__init__(steps)

class NiceLogPipeline(Pipeline):
    def __init__(self, input, output, tail=False):
        stop_event = threading.Event()

        steps = [
            read_transform(input, tail=tail),
            parse_transform(stop_event),
            fanout(
                nice_transform(stop_event),
                passing_report_transform(stop_event),
                fails_report_transform(stop_event),
                totals_report_transform(stop_event),
                version_report_transform(stop_event),
            ),
            fanin(
                "".join
            ),
            write_transform(output),
            stop_transform(stop_event)
        ]
        super(NiceLogPipeline, self).__init__(steps)

class DotsLogPipeline(Pipeline):
    def __init__(self, input, output, tail=False):
        stop_event = threading.Event()

        steps = [
            read_transform(input, tail=tail),
            parse_transform(stop_event),
            fanout(
                dots_transform(stop_event),
                passing_report_transform(stop_event),
                fails_report_transform(stop_event),
                totals_report_transform(stop_event),
                version_report_transform(stop_event),
            ),
            fanin(
                "".join
            ),
            write_transform(output),
            stop_transform(stop_event)
        ]
        super(DotsLogPipeline, self).__init__(steps)

class TotalsReportLogPipeline(Pipeline):
    def __init__(self, input, output):
        stop_event = threading.Event()

        message_types = [Message.TEST] + ResultMessages
        command = f"grep -E '^({'|'.join([str(int(i)) for i in message_types])}),'"

        steps = [
            read_and_filter_transform(input, command=command),
            parse_transform(stop_event),
            fanout(
                totals_report_transform(stop_event),
            ),
            fanin(
                "".join
            ),
            write_transform(output),
            stop_transform(stop_event)
        ]
        super(TotalsReportLogPipeline, self).__init__(steps)

class FailsReportLogPipeline(Pipeline):
    def __init__(self, input, output):
        stop_event = threading.Event()

        message_types = [Message.TEST] + ResultMessages
        command = f"grep -E '^({'|'.join([str(int(i)) for i in message_types])}),'"

        steps = [
            read_and_filter_transform(input, command=command),
            parse_transform(stop_event),
            fanout(
                fails_report_transform(stop_event),
            ),
            fanin(
                "".join
            ),
            write_transform(output),
            stop_transform(stop_event)
        ]
        super(FailsReportLogPipeline, self).__init__(steps)

class PassingReportLogPipeline(Pipeline):
    def __init__(self, input, output):
        stop_event = threading.Event()

        message_types = [Message.TEST] + ResultMessages
        command = f"grep -E '^({'|'.join([str(int(i)) for i in message_types])}),'"

        steps = [
            read_and_filter_transform(input, command=command),
            parse_transform(stop_event),
            fanout(
                passing_report_transform(stop_event),
            ),
            fanin(
                "".join
            ),
            write_transform(output),
            stop_transform(stop_event)
        ]
        super(PassingReportLogPipeline, self).__init__(steps)

class VersionReportLogPipeline(Pipeline):
    def __init__(self, input, output):
        stop_event = threading.Event()

        message_types = [Message.VERSION] + ResultMessages
        command = f"grep -E '^({'|'.join([str(int(i)) for i in message_types])}),'"

        steps = [
            read_and_filter_transform(input, command=command),
            parse_transform(stop_event),
            fanout(
                version_report_transform(stop_event),
            ),
            fanin(
                "".join
            ),
            write_transform(output),
            stop_transform(stop_event)
        ]
        super(VersionReportLogPipeline, self).__init__(steps)

class ResultsLogPipeline(Pipeline):
    def __init__(self, input, results):
        stop_event = threading.Event()
        message_types = [Message.VERSION, Message.TEST] + ResultMessages
        command = f"grep -E '^({'|'.join([str(int(i)) for i in message_types])}),'"

        steps = [
            read_and_filter_transform(input, command=command),
            parse_transform(stop_event),
            results_transform(results, stop_event),
            stop_transform(stop_event)
        ]
        super(ResultsLogPipeline, self).__init__(steps)

class IndexLogPipeline(Pipeline):
    def __init__(self, input, output):
        stop_event = threading.Event()

        steps = [
            read_and_filter_transform(input, command=f"grep -E '^{Message.TEST},'"),
            index_transform(stop_event),
            write_transform(output),
            stop_transform(stop_event)
        ]
        super(IndexLogPipeline, self).__init__(steps)



# Copyright 2019 Vitaliy Zakaznikov (TestFlows Test Framework http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import threading
import testflows._core.cli.arg.type as argtype

from testflows._core.message import Message, ResultMessages
from testflows._core.cli.arg.common import epilog
from testflows._core.cli.arg.common import HelpFormatter
from testflows._core.cli.arg.handlers.handler import Handler as HandlerBase
from testflows._core.transform.log.pipeline import Pipeline as PipelineBase
from testflows._core.transform.log.read_and_filter import transform as read_and_filter_transform
from testflows._core.transform.log.report.results import transform as results_transform
from testflows._core.transform.log.values import transform as values_transform
from testflows._core.transform.log.parse import transform as parse_transform
from testflows._core.transform.log.write import transform as write_transform

from testflows._core.testtype import TestType

def test_details_transform():
    """Transform parsed log line into a short format.
    """
    test = None
    while True:
        line = None
        if test is not None:
            if test["test"].p_type >= TestType.Test:
                line = f"{test['test'].name}\n"
        test = yield line

class Handler(HandlerBase):
    @classmethod
    def add_command(cls, commands):
        parser = commands.add_parser("details", help="details", epilog=epilog(),
            description="Show details.",
            formatter_class=HelpFormatter)

        parser.add_argument("--log", metavar="input", type=argtype.file("r", bufsize=1, encoding="utf-8"),
                nargs="?", help="input log, default: stdin", default="-")
        parser.add_argument("--output", metavar="output", type=argtype.file("w", bufsize=1, encoding="utf-8"),
                nargs="?", help='output, default: stdout', default="-")

        parser.set_defaults(func=cls())

    class FirstStage(PipelineBase):
        def __init__(self, results, input):
            stop_event = threading.Event()
            message_types = [Message.TEST] + ResultMessages
            command = f"grep -E '^({'|'.join([str(int(i)) for i in message_types])}),'"
            steps = [
                read_and_filter_transform(input, command=command),
                parse_transform(stop_event),
                results_transform(results, stop_event),
            ]
            super(Handler.FirstStage, self).__init__(steps)

    class SecondStage(PipelineBase):
        def __init__(self, results, output):
            steps = [
                values_transform(results.values()),
                test_details_transform(),
                write_transform(output)
            ]
            super(Handler.SecondStage, self).__init__(steps)

    def handle(self, args):
        results = {}
        self.FirstStage(results, args.log).run()
        self.SecondStage(results, args.output).run()



# Copyright 2019 Vitaliy Zakaznikov (TestFlows Test Framework http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import inspect
import importlib
import threading

from .exceptions import ResultException
from .serialize import dumps
from .message import Message
from .objects import OK, Fail, Error, Skip, Null
from .objects import XOK, XFail, XError, XNull
from .testtype import TestSubType

#: thread local values
_current_test = {}

def top(value=None, thread=None):
    """Highest level test.
    """
    if thread is None:
        thread = threading.current_thread()
    if _current_test.get(thread.ident) is None:
        _current_test[thread.ident] = {}
    if value is not None:
        _current_test[thread.ident]["main"] = value
    return _current_test[thread.ident].get("main")     

def current(value=None, thread=None):
    """Currently executing test.
    """
    if thread is None:
        thread = threading.current_thread()
    if _current_test.get(thread.ident) is None:
        _current_test[thread.ident] = {}
    if value is not None:
        _current_test[thread.ident]["object"] = value
    return _current_test[thread.ident].get("object")  

def previous(value=None, thread=None):
    """Last executed test.
    """
    if thread is None:
        thread = threading.current_thread()
    if _current_test.get(thread.ident) is None:
        _current_test[thread.ident] = {}
    if value is not None:
        _current_test[thread.ident]["previous"] = value
    return _current_test[thread.ident].get("previous")  

def load(module, test=None):
    """Load test from module path.

    :param module: module path
    :param test: test class or method to load (optional)
    """
    module = importlib.import_module(module)
    if test:
        test = getattr(module, test, None)
    if test is None:
        test = getattr(module, "TestCase", None)
    if test is None:
        test = getattr(module, "TestSuite", None)
    if test is None:
        test = getattr(module, "Test", None)
    return test

def append_path(pathlist, path, *rest, **kwargs):
    """Append path relative to the caller
    to the path list.

    :param pathlist: path list
    :param path: path relative to the caller
    :param *rest: rest of the path
    :param pos: insert at given position,
       default: append to the end of the list
    """
    pos = kwargs.pop("pos", None)
    frame = inspect.currentframe().f_back
    dir = os.path.join(os.path.dirname(os.path.abspath(frame.f_globals["__file__"])), path, *rest)
    if dir not in pathlist:
        if pos is None:
            pathlist.append(dir)
        else:
            pathlist.insert(pos, dir)
    return pathlist

def main(frame=None):
    """Return true if caller is the main module.

    :param frame: caller frame
    """
    if frame is None:
        frame = inspect.currentframe().f_back
    return frame.f_globals["__name__"] == "__main__"

class args(dict):
    pass

def note(message, test=None):
    if test is None:
        test = current()
    test.io.output.note(message)

def debug(message, test=None):
    if test is None:
        test = current()
    test.io.output.debug(message)

def trace(message, test=None):
    if test is None:
        test = current()
    test.io.output.trace(message)

def message(message, test=None):
    if test is None:
        test = current()
    test.io.output.message(Message.NONE, dumps(str(message)))

def value(name, value, test=None):
    if test is None:
        test = current()
    test.io.output.value(name, value)
    return value

def exception(test=None):
    if test is None:
        test = current()
    test.io.output.exception()

def ok(message=None, test=None):
    if test is None:
        test = current()
    test.result = OK(test.name, message)
    raise ResultException(test.result)

def fail(message=None, test=None):
    if test is None:
        test = current()
    test.result = Fail(test.name, message)
    raise ResultException(test.result)

def skip(message=None, test=None):
    if test is None:
        test = current()
    test.result = Skip(test.name, message)
    raise ResultException(test.result)

def error(message=None, test=None):
    if test is None:
        test = current()
    test.result = Error(test.name, message)
    raise ResultException(test.result)

def null(test=None):
    if test is None:
        test = current()
    test.result = Null(test.name)
    raise ResultException(test.result)

def xok(message=None, test=None):
    if test is None:
        test = current()
    test.result = XOK(test.name, message)
    raise ResultException(test.result)

def xfail(message=None, test=None):
    if test is None:
        test = current()
    test.result = XFail(test.name, message)
    raise ResultException(test.result)

def xerror(message=None, test=None):
    if test is None:
        test = current()
    test.result = XError(test.name, message)
    raise ResultException(test.result)

def xnull(test=None):
    if test is None:
        test = current()
    test.result = XNull(test.name)
    raise ResultException(test.result)

def pause(test=None):
    if test is None:
        test = current()
    test.io.output.input("Paused, enter any key to continue...")
    input()

def enter_context(cm, test=None):
    if test is None:
        test = current()
    if not test.caller_test or test.caller_test.subtype != TestSubType.Background:
        raise TypeError("not inside a background test")
    return test.caller_test.stack.enter_context(cm)



# Copyright 2019 Vitaliy Zakaznikov (TestFlows Test Framework http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json

import testflows.settings as settings

from testflows._core.message import MessageMap
from testflows._core.transform.log import message

def transform(stop=None):
    """Transform log line into an index file entry.

    :param stop: stop event, default: None
    """
    prefix = message.RawFormat.prefix
    message_map = message.message_map

    msg = None
    while True:
        if msg is not None:
            try:
                fields = json.loads(f"[{msg}]")
                keyword = fields[prefix.keyword]
                if issubclass(message_map[keyword], message.RawTest):
                    test = message_map[keyword](*fields)
                    msg = f"\"{test.name}\",{test.p_id}\n"
                else:
                    msg = None
            except (IndexError, Exception):
                raise Exception(f"invalid message: {msg}\n")
        msg = yield msg



# Copyright 2019 Vitaliy Zakaznikov (TestFlows Test Framework http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
import subprocess

def transform(file, command, tail=False):
    """Read lines from a file-like object and filter them using
    Unix tail and grep utilities.

    :param file: open file handle
    :pram
    :param tail: tail mode, default: False
    """
    yield None

    process = subprocess.Popen(command, stdin=file, stdout=subprocess.PIPE, shell=True)
    while True:
        line = process.stdout.readline()
        if line == b"":
            if process.poll() is not None:
                break
            else:
                continue
        yield line.decode("utf-8")




# Copyright 2019 Vitaliy Zakaznikov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from hashlib import sha1
from collections import namedtuple

InitArgs = namedtuple("InitArgs", "args kwargs")

def get(a, b):
    """a if not a is None else b.
    """
    return a if not a is None else b

def hash(*s):
    """Calculate standard hash.
    
    :param s: strings
    """
    return sha1(''.join(s).encode("utf-8")).hexdigest()[:32]


class TestObject(object):
    """Base class for all the test objects.
    """
    #: object fields used to represent state of the object
    _fields = ()
    #: defaults for the fields
    _defaults = ()

    def __new__(cls, *args, **kwargs):
        obj = super(TestObject, cls).__new__(cls)
        obj.initargs=InitArgs(
            args=[a for a in args],
            kwargs={k: v for k,v in kwargs.items()})
        return obj

    @property
    def id(self):
        return hash(*[repr(getattr(self, field)) for field in self._fields if field != "id"])

    def __repr__(self):
        """Custom object representation.
        """
        args = ",".join([repr(arg) for arg in self.initargs.args])
        kwargs = ",".join([name + "=" + repr(value) for name, value in self.initargs.kwargs.items()])
        name = self.__class__.__name__
        if args and kwargs:
            args += ","
        return "%s(%s%s)" % (name, args, kwargs)


class TestArg(TestObject):
    """Base class for all test argument object.
    """
    pass

class Table(TestObject, tuple):
    _fields = ("header", "rows", "row_format")
    _defaults = (None, ) * 3
    _row_type_name = "Row"

    def __new__(cls, header=None, rows=None, row_format=None):
        if rows is None:
            rows = []
        if header is None:
            header = ""
        row_type = namedtuple(cls._row_type_name, header)
        obj = tuple.__new__(Table, [row_type(*row) for row in rows])
        obj.initargs=InitArgs(
            args=[header, rows, row_format],
            kwargs={})
        obj.header = header
        obj.row_format = row_format
        obj.rows = obj
        obj.row_type = row_type
        # if no row_formatter is given then use the header as an example
        # or first row, whichever is longer
        if obj.row_format is None:
            sample_row = row_type._fields
            if len(obj) > 0 and len(str(row_type._fields)) < len(str(tuple(obj[0]))):
                sample_row = obj[0]
            obj.row_format = " | ".join([f"%-{max(len(str(c)), 3)}s" for c in sample_row])
        return obj

    def __str__(self):
        """Return markdown-styled table representation.
        """
        s = [self.row_format % self.row_type._fields]
        s.append(self.row_format % tuple(["-" * len(field) for field in self.row_type._fields]))
        s += [self.row_format % row for row in self]
        return "\n".join(s)



# Copyright 2019 Vitaliy Zakaznikov (TestFlows Test Framework http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect

import testflows._core.objects as objects
from collections import namedtuple as namedtuple

from testflows._core.message import MessageMap
from testflows._core.exceptions import TestFlowsError

def namedtuple_with_defaults(*args, defaults=()):
    nt = namedtuple(*args)
    nt.__new__.__defaults__ = defaults
    return nt

class Message(object):
    __slots__ = ()

class InvalidMessageError(TestFlowsError):
    pass

class ProtocolMessage(Message):
    pass

class VersionMessage(Message):
    pass

class InputMessage(Message):
    pass

class NoteMessage(Message):
    pass

class DebugMessage(Message):
    pass

class TraceMessage(Message):
    pass

class NoneMessage(Message):
    pass

class ExceptionMessage(Message):
    pass

class ValueMessage(Message):
    pass

class TestMessage(Message):
    pass

class ResultMessage(Message):
    pass

class OKMessage(ResultMessage):
    pass

class FailMessage(ResultMessage):
    pass

class SkipMessage(ResultMessage):
    pass

class ErrorMessage(ResultMessage):
    pass

class NullMessage(ResultMessage):
    pass

class XOKMessage(ResultMessage):
    pass

class XFailMessage(ResultMessage):
    pass

class XErrorMessage(ResultMessage):
    pass

class XNullMessage(ResultMessage):
    pass

class Prefix(object):
    __slots__ = ()
    fields = "p_keyword p_hash p_num p_type p_subtype p_id p_flags p_cflags p_stream p_time "
    keyword, hash, num, type, subtype, id, flags, cflags, stream, time = list(range(0, 10))

class RawFormat():
    prefix = Prefix

class RawNone(RawFormat, NoneMessage, namedtuple_with_defaults(
        "RawNoneMessage",
        RawFormat.prefix.fields + "message")):
    pass

class RawValue(RawFormat, ValueMessage, namedtuple_with_defaults(
        "RawValueMessage",
        RawFormat.prefix.fields + "name value")):
    pass

class RawException(RawFormat, ExceptionMessage, namedtuple_with_defaults(
        "RawExceptionMessage",
        RawFormat.prefix.fields + "exception")):
    pass

class RawTest(RawFormat, TestMessage, namedtuple_with_defaults(
        "RawTestMessage",
        RawFormat.prefix.fields + "name started flags uid description attributes requirements " +
            "args tags users tickets examples",
        defaults=(None, None, None, [], [], [], [], [], [], None))):

    def __new__(cls, *args):
        args = list(args)
        l = len(args)
        if l > 14 and args[14]: # description
            args[14] = RawDescription(args[14])
        if l > 15 and args[15]: # attributes
            args[15] = [RawAttribute(*attr) for attr in args[15]]
        if l > 16 and args[16]: # requirements
            args[16] = [RawRequirement(*req) for req in args[16]]
        if l > 17 and args[17]: # args
            args[17] = [RawArgument(*arg) for arg in args[17]]
        if l > 18 and args[18]: # tags
            args[18] = [RawTag(*tag) for tag in args[18]]
        if l > 19 and args[19]: # users
            args[19] = [RawUser(*user) for user in args[19]]
        if l > 20 and args[20]:  # tickets
            args[20] = [RawTicket(*ticket) for ticket in args[20]]
        if l > 21 and args[21]:  # examples
            args[21] = RawExamples(*args[21])
        return super(RawTest, cls).__new__(cls, *args)

class RawDescription(namedtuple_with_defaults(
        "RawDescription",
        "description")):
    pass

class RawTag(namedtuple_with_defaults(
        "RawTag",
        " ".join(objects.Tag._fields),
        defaults=objects.Tag._defaults)):
    pass

class RawUser(namedtuple_with_defaults(
        "RawUser",
        " ".join(objects.User._fields),
        defaults=objects.User._defaults)):
    pass

class RawTicket(namedtuple_with_defaults(
        "RawTicket",
        " ".join(objects.Ticket._fields),
        defaults=objects.Ticket._defaults)):
    pass

class RawExamples(namedtuple_with_defaults(
        "RawExamples",
        " ".join(objects.ExamplesTable._fields),
        defaults=objects.ExamplesTable._defaults)):
    pass

class RawArgument(namedtuple_with_defaults(
        "RawArgument",
        " ".join(objects.Argument._fields),
        defaults=objects.Argument._defaults)):
    pass

class RawAttribute(namedtuple_with_defaults(
        "RawAttribute",
        " ".join(objects.Attribute._fields),
        defaults=objects.Attribute._defaults)):
    pass

class RawRequirement(namedtuple_with_defaults(
        "RawRequirement",
        " ".join(objects.Requirement._fields),
        defaults=objects.Requirement._defaults)):
    pass

class RawResultOK(RawFormat, OKMessage, namedtuple_with_defaults(
        "RawResultOKMessage",
        RawFormat.prefix.fields + " ".join(objects.OK._fields),
        defaults=objects.OK._defaults)):
    name = "OK"

class RawResultFail(RawFormat, FailMessage, namedtuple_with_defaults(
        "RawResultFailMessage",
        RawFormat.prefix.fields + " ".join(objects.Fail._fields),
        defaults=objects.Fail._defaults)):
    name = "Fail"

class RawResultSkip(RawFormat, SkipMessage, namedtuple_with_defaults(
        "RawResultSkipMessage",
        RawFormat.prefix.fields + " ".join(objects.Skip._fields),
        defaults=objects.Skip._defaults)):
    name = "Skip"

class RawResultError(RawFormat, ErrorMessage, namedtuple_with_defaults(
        "RawResultErrorMessage",
        RawFormat.prefix.fields + " ".join(objects.Error._fields),
        defaults=objects.Error._defaults)):
    name = "Error"

class RawResultNull(RawFormat, NullMessage, namedtuple_with_defaults(
        "RawResultNullMessage",
        RawFormat.prefix.fields + " ".join(objects.Null._fields),
        defaults=objects.Null._defaults)):
    name = "Null"

class RawResultXOK(RawFormat, XOKMessage, namedtuple_with_defaults(
        "RawResultXOKMessage",
        RawFormat.prefix.fields + " ".join(objects.XOK._fields),
        defaults=objects.XOK._defaults)):
    name = "XOK"

class RawResultXFail(RawFormat, XFailMessage, namedtuple_with_defaults(
        "RawResultXFailMessage",
        RawFormat.prefix.fields + " ".join(objects.XFail._fields),
        defaults=objects.XFail._defaults)):
    name = "XFail"

class RawResultXError(RawFormat, XErrorMessage, namedtuple_with_defaults(
        "RawResultXErrorMessage",
        RawFormat.prefix.fields + " ".join(objects.XError._fields),
        defaults=objects.XError._defaults)):
    name = "XError"

class RawResultXNull(RawFormat, XNullMessage, namedtuple_with_defaults(
        "RawResultXNullMessage",
        RawFormat.prefix.fields + " ".join(objects.XNull._fields),
        defaults=objects.XNull._defaults)):
    name = "XNull"

class RawNote(RawFormat, NoteMessage, namedtuple_with_defaults(
        "RawNoteMessage",
        RawFormat.prefix.fields + "message")):
    pass

class RawTrace(RawFormat, TraceMessage, namedtuple_with_defaults(
        "RawTraceMessage",
        RawFormat.prefix.fields + "message")):
    pass

class RawDebug(RawFormat, DebugMessage, namedtuple_with_defaults(
        "RawDebugMessage",
        RawFormat.prefix.fields + "message")):
    pass

class RawInput(RawFormat, InputMessage, namedtuple_with_defaults(
        "RawInputMessage",
        RawFormat.prefix.fields + "message")):
    pass

class RawProtocol(RawFormat, ProtocolMessage, namedtuple_with_defaults(
        "RawProtocolMessage",
        RawFormat.prefix.fields + "version")):
    pass

class RawVersion(RawFormat, VersionMessage, namedtuple_with_defaults(
        "RawVersionMessage",
        RawFormat.prefix.fields + "version")):
    pass

message_map = MessageMap(
    RawNone, # NONE
    RawTest, # TEST
    RawResultNull, # NULL
    RawResultOK, # OK
    RawResultFail, # FAIL
    RawResultSkip, # SKIP
    RawResultError, # ERROR
    RawException, # EXCEPTION
    RawValue, # VALUE
    RawNote, # NOTE
    RawDebug, # DEBUG
    RawTrace, # TRACE
    RawResultXOK, #XOK
    RawResultXFail, # XFAIL
    RawResultXError, # XERROR
    RawResultXNull, # XNULL
    RawProtocol, # PROTOCOL
    RawInput, # INPUT
    RawVersion, # VERSION
)


# Copyright 2019 Vitaliy Zakaznikov, TestFlows Test Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re

from testflows._core.contrib.arpeggio import RegExMatch as _
from testflows._core.contrib.arpeggio import OneOrMore, ZeroOrMore, EOF, Optional, Not
from testflows._core.contrib.arpeggio import ParserPython as PEGParser
from testflows._core.contrib.arpeggio import PTNodeVisitor, visit_parse_tree

template = """
%(pyname)s = Requirement(
        name='%(name)s',
        version='%(version)s',
        priority=%(priority)s,
        group=%(group)s,
        type=%(type)s,
        uid=%(uid)s,
        description=%(description)s,
        link=%(link)s
    )

"""

class Visitor(PTNodeVisitor):
    def __init__(self, *args, **kwargs):
        self.output = (
            "# These are auto generated requirements from an SRS document.\n"
            "# Do not edit by hand but re-generate instead\n"
            "# using \"tfs requirement generate\" command.\n"
            "#\n"
            "from testflows.core import Requirement\n\n"
            )
        self.pyname_fmt = re.compile(r"[^a-zA-Z0-9]")
        super(Visitor, self).__init__(*args, **kwargs)

    def visit_line(self, node, children):
        pass

    def visit_requirement(self, node, children):
        name = node.requirement_heading.requirement_name.value
        pyname = re.sub(r"_+", "_", self.pyname_fmt.sub("_", name))
        description = None
        group = None
        priority = None
        type = None
        uid = None
        link = None

        try:
            description = "\n".join([f'{"":8}{repr(line.value)}' for lines in node.requirement_description for line in lines])
            description = f"(\n{description}\n{'':8})"
        except:
            pass
        try:
            priority = node.priority.word
        except:
            pass
        try:
            group = f"\"{node.group.word}\""
        except:
            pass
        try:
            type = f"\"{node.type.word}\""
        except:
            pass
        try:
            uid = f"\"{node.uid.word}\""
        except:
            pass

        self.output += template.lstrip() % {
            "pyname": pyname,
            "name": node.requirement_heading.requirement_name.value,
            "version": node.version.word,
            "description": str(description),
            "priority": str(priority),
            "group": str(group),
            "type": str(type),
            "uid": str(uid),
            "link": str(link)
        }

    def visit_document(self, node, children):
        return self.output.rstrip() + "\n"

def Parser():
    """Returns markdown requirements parser.
    """
    def line():
        return _(r"[^\n]*\n")

    def not_heading():
        return Not(heading)

    def heading():
        return [
            (_(r"\s?\s?\s?#+\s+"), heading_name, _(r"\n?")),
            (heading_name, _(r"[-=]+\n?"))
        ]

    def requirement_heading():
        return [
            (_(r"\s?\s?\s?#+\s+"), requirement_name, _(r"\n?")),
            (requirement_name, _(r"[-=]+\n?"))
        ]

    def heading_name():
        return _(r"[^\n]+")

    def requirement_name():
        return _(r"RQ\.[^\n]+")

    def requirement_description():
        return ZeroOrMore((not_heading, line))

    def word():
        return _(r"[^\s]+")

    def version():
        return _(r"\s*version:\s*"), word

    def priority():
        return _(r"\s*priority:\s*"), word

    def type():
        return _(r"\s*type:\s*"), word

    def group():
        return _(r"\s*group:\s*"), word

    def uid():
        return _(r"\s*uid:\s*"), word

    def requirement():
        return requirement_heading, version, ZeroOrMore([priority, type, group, uid]), Optional(requirement_description), _(r"\n?")

    def document():
        return Optional(OneOrMore([requirement, heading, line])), EOF

    return PEGParser(document)


def generate(source, destination):
    """Generate requirements from markdown source.

    :param source: source file-like object
    :param destination: destination file-like object
    """
    parser = Parser()
    source_data = source.read()
    tree = parser.parse(source_data)
    destination_data = visit_parse_tree(tree, Visitor())
    if destination_data:
        destination.write(destination_data)



# Copyright 2019 Vitaliy Zakaznikov (TestFlows Test Framework http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import testflows.settings as settings

from testflows._core.flags import Flags, SKIP
from testflows._core.transform.log import message
from testflows._core.cli.colors import color

width = 70
count = 0

def color_result(result):
    if result.startswith("X"):
        return color(".", "blue", attrs=["bold"])
    elif result == "OK":
        return color(".", "green", attrs=["bold"])
    elif result == "Skip":
        return color("-", "white", attrs=["dim"])
    # Error, Fail, Null
    elif result == "Error":
        return color("E", "yellow", attrs=["bold"])
    elif result == "Fail":
         return color("F", "red", attrs=["bold"])
    elif result == "Null":
        return color("N", "magenta", attrs=["bold"])
    else:
        raise ValueError(f"unknown result {result}")

def format_result(msg, result):
    global count
    flags = Flags(msg.p_flags)
    if flags & SKIP and settings.show_skipped is False:
        return

    count += 1
    _result = f"{color_result(result)}"
    # wrap if we hit max width
    if count >= width:
        count = 0
        _result += "\n"

    return _result

formatters = {
    message.RawResultOK: (format_result, f"OK"),
    message.RawResultFail: (format_result, f"Fail"),
    message.RawResultError: (format_result, f"Error"),
    message.RawResultSkip: (format_result, f"Skip"),
    message.RawResultNull: (format_result, f"Null"),
    message.RawResultXOK: (format_result, f"XOK"),
    message.RawResultXFail: (format_result, f"XFail"),
    message.RawResultXError: (format_result, f"XError"),
    message.RawResultXNull: (format_result, f"XNull")
}

def transform(stop_event):
    """Transform parsed log line into a short format.
    """
    line = None
    while True:
        if line is not None:
            formatter = formatters.get(type(line), None)
            if formatter:
                line = formatter[0](line, *formatter[1:])
                n = 0
            else:
                line = None

        if stop_event.is_set():
            if line is None:
                line = ""
            line += "\n"
        line = yield line



# Copyright 2019 Vitaliy Zakaznikov (TestFlows Test Framework http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from testflows._core.transform.log import message
from testflows._core.transform.log.report.totals import Counts, all_counts
from testflows._core.transform.log.report.totals import format_test as process_test_counts
from testflows._core.transform.log.report.totals import format_result as process_result_counts

def process_test(msg, results):
    results["tests"][msg.name] = {"test": msg}
    process_test_counts(msg, results["counts"])

def process_result(msg, results):
    results["tests"][msg.test]["result"] = msg
    process_result_counts(msg, results["counts"])

def process_version(msg, results):
    results["version"] = msg.version
    results["started"] = msg.p_time

processors = {
    message.RawVersion: process_version,
    message.RawTest: process_test,
    message.RawResultOK: process_result,
    message.RawResultFail: process_result,
    message.RawResultError: process_result,
    message.RawResultSkip: process_result,
    message.RawResultNull: process_result,
    message.RawResultXOK: process_result,
    message.RawResultXFail: process_result,
    message.RawResultXError: process_result,
    message.RawResultXNull: process_result,
}

def transform(results, stop_event):
    """Transform parsed log line into a short format.
    """
    if results.get("tests") is None:
        results["tests"] = {}

    if results.get("counts") is None:
        results["counts"] = all_counts()

    line = None
    while True:
        if line is not None:
            processor = processors.get(type(line), None)
            if processor:
                processor(line, results)

        line = yield line


# Copyright 2019 Vitaliy Zakaznikov (TestFlows Test Framework http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from testflows._core.cli.arg.common import epilog
from testflows._core.cli.arg.common import HelpFormatter
from testflows._core.cli.arg.handlers.handler import Handler as HandlerBase
from testflows._core.cli.arg.handlers.show.test.handler import Handler as test_handler
from testflows._core.cli.arg.handlers.show.tests.handler import Handler as tests_handler

class Handler(HandlerBase):
    @classmethod
    def add_command(cls, commands):
        parser = commands.add_parser("show", help="show test data", epilog=epilog(),
            description="Show test data.",
            formatter_class=HelpFormatter)

        show_commands = parser.add_subparsers(title="commands", metavar="command",
            description=None, help=None)
        show_commands.required = True
        test_handler.add_command(show_commands)
        tests_handler.add_command(show_commands)
